@@ -9,6 +9,7 @@ from . import genera
 from ._version import __version__
 
 from .genera import sanirule, genus_to_module, obtain_genus
+from .pythlib.lowlevel import regstring
 
 __all__ = ['lifelib_dir', 'compile_rules', 'load_rules', 'reset_tree', 'set_rules',
            'sanirule', 'run_tests', 'add_cygdir', 'install_cygwin']
@@ -106,8 +107,6 @@ def install_cygwin(root_directory=None, packages=['python2', 'python2-numpy', 'p
 
 def verify_installation():
 
-    from .pythlib.lowlevel import regstring
-
     print("Verifying installation...")
     subprocess.check_call(get_local_bash() + ['echo', 'SUCCESS'])
 
@@ -158,7 +157,7 @@ def run_tests():
 
 
 def obtain_arguments(soname, standard="C++11", optimisations=["-march=native", "-O3"],
-                    options=["-Wall", "-Wextra"], **unused_kwargs):
+                    options=["-Wall", "-Wextra", "-fdiagnostics-color=always"], **unused_kwargs):
 
     so_args = ["-fPIC", "-shared", "-o", soname, "lifelib.cpp"]
 
@@ -263,8 +262,8 @@ def compile_rules(*rules, **kwargs):
                     pbar.update(1)
 
             if kwargs.get('compiler_output', (status != 0)):
-                sys.stderr.write("Compiler stdout: %s\n" % g_stdout)
-                sys.stderr.write("Compiler stderr: %s\n" % g_stderr)
+                sys.stderr.write("Compiler stdout: %s\n" % regstring(g_stdout))
+                sys.stderr.write("Compiler stderr: %s\n" % regstring(g_stderr))
 
             if (status != 0):
                 raise RuntimeError("g++ returned nonzero exit code %d" % status)
