@@ -453,7 +453,12 @@ namespace apg {
                         // Separating standard spaceships is considerably
                         // faster and more reliable with sss; only fall
                         // back on pbb if this fails:
-                        elements = sss(cl2);
+                        if ((repr == "xq4_27deee6zwsighhga") || (repr == "xq4_06eeed72zaghgis")) {
+                            // bypass sss()
+                            elements.push_back(repr);
+                        } else {
+                            elements = sss(cl2);
+                        }
                     }
                     if (elements.size() == 0) { elements = pseudoBangBang(cl2, 0, recurse); }
                 } else if (recurse && (!b0)) {
