@@ -85,6 +85,27 @@ extern "C"
         out << footer << std::endl;
     }
 
+    void LoadTimelineMC(void* lt, const char *filename, int maxframes, void* ptr) {
+        auto lab = reinterpret_cast<apg::lifetree_abstract<uint32_t>*>(lt);
+        std::vector<apg::hypernode<uint32_t> > hvec;
+        std::string rule = "";
+        lab->load_macrocell(filename, rule, &hvec);
+        apg::pattern** patarray = reinterpret_cast<apg::pattern**>(ptr);
+
+        int elf = 0;
+        for (auto it = hvec.begin(); it != hvec.end(); ++it) {
+            if (elf == maxframes) { break; }
+            if (it->index != 0) {
+                patarray[elf] = new(std::nothrow) apg::pattern(lab, (*it), rule);
+                elf += 1;
+            }
+        }
+
+        while (elf < maxframes) {
+            patarray[elf] = 0; elf += 1;
+        }
+    }
+
     void SaveTimelineMC(void* lt, const char *filename, const char *header, const char* footer, void *ptr, int exponent) {
 
         auto lab = reinterpret_cast<apg::lifetree_abstract<uint32_t>*>(lt);
