@@ -443,8 +443,8 @@ namespace apg {
             */
 
             std::string line; // line read from file
-            std::vector<nicearray<uint64_t, N> > pleaves; // partial leaves
-            nicearray<uint64_t, N> zeropleaf = {0ull}; // empty partial leaf
+            std::vector<nicearray<uint64_t, N+1> > pleaves; // partial leaves
+            nicearray<uint64_t, N+1> zeropleaf = {0ull}; // empty partial leaf
             pleaves.push_back(zeropleaf); // zero means zero
 
             uint32_t log2size = 0; // log2(size) of most recent node
@@ -501,7 +501,7 @@ namespace apg {
                 } else if (reading_rle) {
                     rlestream << line;
                 } else {
-                    nicearray<uint64_t, N> pleaf = {0ull};
+                    nicearray<uint64_t, N+1> pleaf = {0ull};
                     if (line[0] == '.' || line[0] == '*' || line[0] == '$') {
                         uint64_t lm = 1;
                         uint64_t pl = 0;
@@ -1249,9 +1249,9 @@ namespace apg {
             }
         }
 
-        hypernode<I> boolean_recurse(hypernode<I> lnode, hypernode<I> rnode, int operation,
-                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap,
-                                        std::vector<hypernode<I> > *solids) {
+        template<typename MemMap>
+        hypernode<I> boolean_recursei(hypernode<I> lnode, hypernode<I> rnode, int operation,
+                                        MemMap *memmap, std::vector<hypernode<I> > *solids) {
             /*
             *   0 = and
             *   1 = or
@@ -1272,7 +1272,7 @@ namespace apg {
                     return lnode;
                 }
             } else if ((rnode.index2 != 0) || (lnode.index2 != 0)) {
-                return boolean_recurse(breach(lnode), breach(rnode), operation, memmap, solids);
+                return boolean_recursei(breach(lnode), breach(rnode), operation, memmap, solids);
             } else if ((solids != 0) && (lnode.index == (*solids)[lnode.depth].index)) {
                 if (operation == 0) { return rnode; } else { return lnode; }
             } else if ((solids != 0) && (rnode.index == (*solids)[rnode.depth].index)) {
@@ -1288,13 +1288,13 @@ namespace apg {
                     // Nonleaf node:
                     kiventry<nicearray<I, 4>, I, J >* lptr = ind2ptr_nonleaf(lnode.depth, lnode.index);
                     kiventry<nicearray<I, 4>, I, J >* rptr = ind2ptr_nonleaf(rnode.depth, rnode.index);
-                    hypernode<I> ytl = boolean_recurse(hypernode<I>(lptr->key.x[0], lnode.depth-1), hypernode<I>(rptr->key.x[0], rnode.depth-1),
+                    hypernode<I> ytl = boolean_recursei(hypernode<I>(lptr->key.x[0], lnode.depth-1), hypernode<I>(rptr->key.x[0], rnode.depth-1),
                                                         operation, memmap, solids);
-                    hypernode<I> ytr = boolean_recurse(hypernode<I>(lptr->key.x[1], lnode.depth-1), hypernode<I>(rptr->key.x[1], rnode.depth-1),
+                    hypernode<I> ytr = boolean_recursei(hypernode<I>(lptr->key.x[1], lnode.depth-1), hypernode<I>(rptr->key.x[1], rnode.depth-1),
                                                         operation, memmap, solids);
-                    hypernode<I> ybl = boolean_recurse(hypernode<I>(lptr->key.x[2], lnode.depth-1), hypernode<I>(rptr->key.x[2], rnode.depth-1),
+                    hypernode<I> ybl = boolean_recursei(hypernode<I>(lptr->key.x[2], lnode.depth-1), hypernode<I>(rptr->key.x[2], rnode.depth-1),
                                                         operation, memmap, solids);
-                    hypernode<I> ybr = boolean_recurse(hypernode<I>(lptr->key.x[3], lnode.depth-1), hypernode<I>(rptr->key.x[3], rnode.depth-1),
+                    hypernode<I> ybr = boolean_recursei(hypernode<I>(lptr->key.x[3], lnode.depth-1), hypernode<I>(rptr->key.x[3], rnode.depth-1),
                                                         operation, memmap, solids);
                     nicearray<I, 4> cc = {ytl.index, ytr.index, ybl.index, ybr.index};
                     hypernode<I> xcc = make_nonleaf_hn(lnode.depth, cc);
@@ -1329,8 +1329,8 @@ namespace apg {
             }
         }
 
-        hypernode<I> boolean_recurse(hypernode<I> lnode, hypernode<I> rnode, int operation,
-                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap) {
+        template<typename MemMap>
+        hypernode<I> boolean_recursei(hypernode<I> lnode, hypernode<I> rnode, int operation, MemMap* memmap) {
 
             std::vector<hypernode<I> > solids;
 
@@ -1344,7 +1344,13 @@ namespace apg {
                 solids.push_back(hnode);
             }
 
-            return boolean_recurse(lnode, rnode, operation, memmap, ((operation <= 1) ? (&solids) : 0));
+            return boolean_recursei(lnode, rnode, operation, memmap, ((operation <= 1) ? (&solids) : 0));
+
+        }
+
+        hypernode<I> boolean_recurse(hypernode<I> lnode, hypernode<I> rnode, int operation, std::map<std::pair<std::pair<I, I>, uint32_t>, I>* memmap) {
+
+            return boolean_recursei(lnode, rnode, operation, memmap);
 
         }
 
@@ -1399,10 +1405,88 @@ namespace apg {
 
         }
 
-        hypernode<I> convolve_recurse(hypernode<I> lnode, hypernode<I> rnode, bool exclusive,
+        template<typename MemMap>
+        hypernode<I> matmul_recursei(hypernode<I> lnode, hypernode<I> rnode, bool exclusive, MemMap *memmap1, MemMap *memmap2) {
+
+            auto it = memmap2->find(std::make_pair(std::make_pair(lnode.index, rnode.index), lnode.depth));
+
+            if (lnode.index == 0) {
+                return lnode;
+            } else if (rnode.index == 0) {
+                return rnode;
+            } else if (it != memmap2->end()) {
+                return hypernode<I>(it->second, lnode.depth);
+            } else if (lnode.depth >= 1) {
+                // Nonleaf node:
+                kiventry<nicearray<I, 4>, I, J >* lptr = ind2ptr_nonleaf(lnode.depth, lnode.index);
+                kiventry<nicearray<I, 4>, I, J >* rptr = ind2ptr_nonleaf(rnode.depth, rnode.index);
+
+                // Extract submatrices:
+                hypernode<I> a11(lptr->key.x[0], lnode.depth - 1);
+                hypernode<I> a12(lptr->key.x[1], lnode.depth - 1);
+                hypernode<I> a21(lptr->key.x[2], lnode.depth - 1);
+                hypernode<I> a22(lptr->key.x[3], lnode.depth - 1);
+
+                hypernode<I> b11(rptr->key.x[0], rnode.depth - 1);
+                hypernode<I> b12(rptr->key.x[1], rnode.depth - 1);
+                hypernode<I> b21(rptr->key.x[2], rnode.depth - 1);
+                hypernode<I> b22(rptr->key.x[3], rnode.depth - 1);
+
+                // The bool 'exclusive' determines whether the operation is OR or XOR:
+                int op = exclusive + 1;
+
+                // TODO use Strassen here when exclusive == true:
+                hypernode<I> c11 = boolean_recursei(matmul_recursei(a11, b11, exclusive, memmap1, memmap2),
+                                                    matmul_recursei(a12, b21, exclusive, memmap1, memmap2), op, memmap1, 0);
+                hypernode<I> c12 = boolean_recursei(matmul_recursei(a11, b12, exclusive, memmap1, memmap2),
+                                                    matmul_recursei(a12, b22, exclusive, memmap1, memmap2), op, memmap1, 0);
+                hypernode<I> c21 = boolean_recursei(matmul_recursei(a21, b11, exclusive, memmap1, memmap2),
+                                                    matmul_recursei(a22, b21, exclusive, memmap1, memmap2), op, memmap1, 0);
+                hypernode<I> c22 = boolean_recursei(matmul_recursei(a21, b12, exclusive, memmap1, memmap2),
+                                                    matmul_recursei(a22, b22, exclusive, memmap1, memmap2), op, memmap1, 0);
+
+                nicearray<I, 4> cc = {c11.index, c12.index, c21.index, c22.index};
+                hypernode<I> xcc = make_nonleaf_hn(lnode.depth, cc);
+                memmap2->emplace(std::make_pair(std::make_pair(lnode.index, rnode.index), lnode.depth), xcc.index);
+                return xcc;
+            } else {
+                // Base case:
+                kiventry<nicearray<uint64_t, 4*N>, I, J >* lptr = ind2ptr_leaf(lnode.index);
+                kiventry<nicearray<uint64_t, 4*N>, I, J >* rptr = ind2ptr_leaf(rnode.index);
+                nicearray<uint64_t, 4*N> outleaf = {0ull};
+                if (exclusive) {
+                    for (int i = 0; i < 4*N; i += 4) {
+                        outleaf.x[i+0] = mxor(lptr->key.x[i+0], rptr->key.x[i+0]) ^ mxor(lptr->key.x[i+1], rptr->key.x[i+2]);
+                        outleaf.x[i+1] = mxor(lptr->key.x[i+0], rptr->key.x[i+1]) ^ mxor(lptr->key.x[i+1], rptr->key.x[i+3]);
+                        outleaf.x[i+2] = mxor(lptr->key.x[i+2], rptr->key.x[i+0]) ^ mxor(lptr->key.x[i+3], rptr->key.x[i+2]);
+                        outleaf.x[i+3] = mxor(lptr->key.x[i+2], rptr->key.x[i+1]) ^ mxor(lptr->key.x[i+3], rptr->key.x[i+3]);
+                    }
+                } else {
+                    for (int i = 0; i < 4*N; i += 4) {
+                        outleaf.x[i+0] = mor(lptr->key.x[i+0], rptr->key.x[i+0]) | mor(lptr->key.x[i+1], rptr->key.x[i+2]);
+                        outleaf.x[i+1] = mor(lptr->key.x[i+0], rptr->key.x[i+1]) | mor(lptr->key.x[i+1], rptr->key.x[i+3]);
+                        outleaf.x[i+2] = mor(lptr->key.x[i+2], rptr->key.x[i+0]) | mor(lptr->key.x[i+3], rptr->key.x[i+2]);
+                        outleaf.x[i+3] = mor(lptr->key.x[i+2], rptr->key.x[i+1]) | mor(lptr->key.x[i+3], rptr->key.x[i+3]);
+                    }
+                }
+                I res = make_leaf(outleaf);
+                memmap2->emplace(std::make_pair(std::make_pair(lnode.index, rnode.index), 0), res);
+                return hypernode<I>(res, 0);
+            }
+        }
+
+        hypernode<I> matmul_recurse(hypernode<I> lnode, hypernode<I> rnode, bool exclusive,
                                         std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap1,
                                         std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap2) {
 
+            return matmul_recursei(lnode, rnode, exclusive, memmap1, memmap2);
+
+        }
+
+        template<typename MemMap>
+        hypernode<I> convolve_recursei(hypernode<I> lnode, hypernode<I> rnode, bool exclusive,
+                                        MemMap *memmap1, MemMap *memmap2) {
+
             auto it = memmap2->find(std::make_pair(std::make_pair(lnode.index, rnode.index), lnode.depth));
 
             if (lnode.index == 0) {
@@ -1427,7 +1511,7 @@ namespace apg {
 
                 for (int i = 0; i < 4; i++) {
                     for (int j = 0; j < 4; j++) {
-                        hypernode<I> q = convolve_recurse(hypernode<I>(lptr->key.x[i], lnode.depth-1),
+                        hypernode<I> q = convolve_recursei(hypernode<I>(lptr->key.x[i], lnode.depth-1),
                                                           hypernode<I>(rptr->key.x[j], rnode.depth-1),
                                                           exclusive, memmap1, memmap2);
                         if (q.index) {
@@ -1438,7 +1522,7 @@ namespace apg {
                                 int b = (i & 2) + (j & 2) + (k & 2);
                                 int c = ((b & 4) | (a & 2)) >> 1;
                                 int d = ((b & 2) | (a & 1));
-                                pconvs[c][d] = boolean_recurse(pconvs[c][d], hijk, op, memmap1, 0);
+                                pconvs[c][d] = boolean_recursei(pconvs[c][d], hijk, op, memmap1, 0);
                             }
                         }
                     }
@@ -1510,6 +1594,14 @@ namespace apg {
 
         }
 
+        hypernode<I> convolve_recurse(hypernode<I> lnode, hypernode<I> rnode, bool exclusive,
+                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap1,
+                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap2) {
+
+            return convolve_recursei(lnode, rnode, exclusive, memmap1, memmap2);
+
+        }
+
         hypernode<I> solid(uint32_t depth) {
             nicearray<uint64_t, 4*N> outleaf;
             std::memset(outleaf.x, 255, 32*N);
