@@ -196,6 +196,26 @@ namespace apg {
             return hnode;
         }
 
+        virtual hypernode<I> matmul_recurse(hypernode<I> lnode, hypernode<I> rnode, bool exclusive,
+                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap1,
+                                        std::map<std::pair<std::pair<I, I>, uint32_t>, I> *memmap2) = 0;
+
+        hypernode<I> matmul_recurse(hypernode<I> lnode, hypernode<I> rnode, bool exclusive) {
+            std::map<std::pair<std::pair<I, I>, uint32_t>, I> memmap1;
+            std::map<std::pair<std::pair<I, I>, uint32_t>, I> memmap2;
+            return matmul_recurse(lnode, rnode, exclusive, &memmap1, &memmap2);
+        }
+
+        hypernode<I> matmul_universe(hypernode<I> lnode, hypernode<I> rnode, bool exclusive) {
+            hypernode<I> lx = lnode;
+            hypernode<I> rx = rnode;
+            while (lx.depth < rx.depth) { lx = pyramid_up(lx); }
+            while (lx.depth > rx.depth) { rx = pyramid_up(rx); }
+            hypernode<I> hnode = matmul_recurse(lx, rx, exclusive);
+            hnode = pyramid_down(hnode);
+            return hnode;
+        }
+
         virtual uint64_t digest_recurse(hypernode<I> hnode,
                                         std::map<std::pair<I, uint32_t>, uint64_t> *memmap) = 0;
 
@@ -239,6 +259,10 @@ namespace apg {
                 return convolve_universe(lnode, rnode, false);
             } else if (operation == 6) {
                 return convolve_universe(lnode, rnode, true);
+            } else if (operation == 7) {
+                return matmul_universe(lnode, rnode, true);
+            } else if (operation == 8) {
+                return matmul_universe(lnode, rnode, false);
             }
 
             hypernode<I> lx = lnode;
